/***************************************

  X Web Teach       version 1.0

  Alan Braverman
  Software Development Group
  National Center for Supercomputing Applications
  alanb@ncsa.uiuc.edu

****************************************/

#include <X11/Intrinsic.h>      /* X includes */
#include <Xm/Xm.h>              /* Motif includes */
#include <Xm/PushB.h> 
#include <Xm/Form.h>
#include <Xm/Text.h>
#include <Xm/Label.h>
#include <Xm/List.h>
#include <Xm/ToggleB.h>
#include <Xm/Separator.h>
#include <Xm/FileSB.h>
#include <Xm/MessageB.h>
#include <stdio.h>              /* C includes */
#include <stdlib.h>
#include "libcci/cci.h"            /* CCI includes */

#define MAX_STUDENTS 255        /* 16 bytes per empty student, no biggie */
#define VERSION "1.0"
#define HELP_URL "http://www.ncsa.uiuc.edu/SDG/Software/Mosaic/Docs/x-web-teach.html"

static char *default_resources[] = 
{
  "XWebTeach*Background: grey70",
  "XWebTeach*Foreground: blue",
  "XWebTeach*XmList*Foreground: black",
  "XWebTeach*XmText*Foreground: black",
  "XWebTeach*XmToggleButton*SelectColor: yellow",
  "XWebTeach*XmLabel*fontList:  -adobe-times-bold-r-normal-*-17-*-*-*-*-*-*-*",
  "XWebTeach*fontList: -adobe-times-medium-r-normal-*-17-*-*-*-*-*-*-*",
  NULL,
};

/* gui globals */
XtAppContext    app;
Widget          toplevel, form_widget,                 /* top level widgets */
                setup_button, edit_button,              
                load_button, save_button, toggle,
                help_button, quit_button, sep,
                add_button, remove_button,
                teacher_field, title_field, url_field,  
                teacher_label, student_label,           
                title_label, url_label,
                student_list, student_list_parent;
Widget          setup_host_field, setup_port_field, /* setup dialog widgets  */
                setup_host_label, setup_port_label,
                setup_form, setup_parent, setup_top_label, setup_pot_label,
                setup_ok_button, setup_cancel_button;
Widget          edit_host_field, edit_port_field, /* edit dialog widgets  */
                edit_host_label, edit_port_label,
                edit_form, edit_parent,
                edit_ok_button, edit_cancel_button;
Widget          add_host_field, add_port_field, /* add dialog widgets  */
                add_host_label, add_port_label,
                add_form, add_parent,
                add_ok_button, add_cancel_button;
Widget          load_dialog, save_dialog, load_parent_widget, 
                save_parent_widget;
XtPopdownIDRec  setup_idrec, edit_idrec, add_idrec, load_idrec, save_idrec;
char            dummy[256];
XmString        *xmstr_student_list;
short           geo_x, geo_y;

/* global data */
char            *teach_host=NULL, *old_teacher_host_string=NULL;
char            *teach_port=NULL, *old_teacher_port_string=NULL;
char            *student_host[MAX_STUDENTS];
char            *student_port[MAX_STUDENTS];
int             num_students=0;
int             current_selection= -1;
int             current_edit= -1;
int             all_greyed_out = 0, message_grey_flag;

/* CCI globals */
MCCIPort        cci_teach_port=NULL;
MCCIPort        cci_student_port[MAX_STUDENTS];
XtInputId       student_input_id[MAX_STUDENTS];
XtInputId       teacher_input_id;
int             broadcasting=0;

/* message globals */
char *help_message[]=
{
  "X Web Teach       version 1.0",
  "",
  "X Web Teach is a program that interacts with multiple sessions",
  "of NCSA Mosaic for the X Window System, running on multiple ",
  "machines. X Web Teach uses a built-in Common Client Interface (CCI)",
  "which allows one Mosaic session to act as a teacher, and multiple",
  "other Mosaic sessions to act as students.  When the teacher",
  "access a new document on the Web, the students follow along.",
  " ",
  "More help is available at:",
  "http://www.ncsa.uiuc.edu/SDG/Software/Mosaic/Docs/x-web-teach.html",
  " ",
  "Use the HELP button below to go there now.",
  " ",
  "(You must have broadcasting enabled, or have Mosaic installed and",
  "have named it 'Mosaic', in order to perform this trick.)",
  "--",
};

char *no_connect_msg[]=
{ 
  "Could not connect to teacher host machine at specified",
  "port number.",
  "--",
};

char *no_teach_host_msg[]=
{ 
  "A teacher machine host and port must be specified.",
  "Use the SETUP button to enter a teacher host machine",
  "and port number.",
  "--",
};

char *add_error_msg[]=
{  
  "You must specify both a hostname and a port number.", 
  "--",
};

char *no_students_msg[]=
{ 
  "No student machines have been entered.", 
  "Use the ADD button or load a saved configuration.",
  "--",
};

char *no_selected_msg[]=
{ 
  "No student entries have been selected.", 
  "--",
};

char *already_used_msg[]=
{
  "That hostname and port number are already listed.",
  "--",
};

/* ---------------------------- ugh ----------------------------------- */
#if defined(ultrix)
char *strdup (char *str)
{
  char *dup;

  dup = (char *)malloc (strlen (str) + 1);
  dup = strcpy (dup, str);

  return dup;
}
#endif


/* ------------------ grey everything out or back on ------------------ */
void xwt_grey_out_all(Boolean state)
{
  if (state==True)
    all_greyed_out=0;
  else
    all_greyed_out=1;
  XtSetSensitive(add_button, state);
  XtSetSensitive(remove_button, state);
  XtSetSensitive(edit_button, state);
  XtSetSensitive(setup_button, state);
  XtSetSensitive(help_button, state);
  XtSetSensitive(toggle, state);
}

/* --------------------- message dialogs -------------------------- */

XmString xwt_str_array_to_xmstr(char *normal_string[]) /* from Young */
{
  XmString return_string;
  int i;

  return_string = (XmString) NULL;
  i=0;
  do
    {
      if (i>0)
	return_string=XmStringConcat(return_string, XmStringSeparatorCreate());
      return_string=XmStringConcat(return_string, 
				   XmStringCreate(normal_string[i], 
						  XmSTRING_DEFAULT_CHARSET));
      i++;
    } while (strcmp(normal_string[i], "--")!=0);
  return return_string;
}

void xwt_message_ok_cb(Widget w, caddr_t client_data, 
		       XmAnyCallbackStruct call_data)
{ 
  if (message_grey_flag==0)
    xwt_grey_out_all(TRUE);
  XtUnmanageChild(w);
}

void xwt_more_help_cb(Widget w, caddr_t client_data, 
		       XmAnyCallbackStruct call_data)
{
  int tmp_broadcast_flag;
  char sys_help_command[256];
  
  if (MCCIIsConnected(cci_teach_port))
    {
      tmp_broadcast_flag=broadcasting;
      broadcasting=0;
      MCCIGet(cci_teach_port, HELP_URL, MCCI_DEFAULT, MCCI_ABSOLUTE, 0);
      broadcasting=tmp_broadcast_flag;
    }
  else
    {
      sprintf(sys_help_command, "Mosaic -home %s &", HELP_URL);
      system(sys_help_command);
    }
}
  
void xwt_message_dialog(Widget parent, char *name, char *message[], int grey)
{
  Widget   dialog_box;
  XmString xmstr_mes;
  Arg      wargs[5];
  int      n=0;
  char     fullname[50];

  if (all_greyed_out==1)
    message_grey_flag=1;
  else
    message_grey_flag=0;
  if (grey==1)
    xwt_grey_out_all(FALSE);
  dialog_box = XmCreateMessageDialog(parent, "message", NULL, 0);
  XtUnmanageChild(XmMessageBoxGetChild(dialog_box, XmDIALOG_CANCEL_BUTTON));
  if (strcmp(name, "Help")!=0)
    XtUnmanageChild(XmMessageBoxGetChild(dialog_box, XmDIALOG_HELP_BUTTON));
  else
    XtAddCallback(dialog_box, XmNhelpCallback, xwt_more_help_cb);
  XtAddCallback(dialog_box, XmNokCallback, xwt_message_ok_cb, grey);
  xmstr_mes=xwt_str_array_to_xmstr(message);
  XtSetArg(wargs[n], XmNmessageString, xmstr_mes); n++;
  XtSetValues(dialog_box, wargs, n);

  strcpy(fullname, "X Web Teach: ");
  strcat(fullname, name);
  n=0;                                 /* set the title of the dialog box */
  XtSetArg(wargs[n], XmNtitle, fullname); n++;
  XtSetValues(XtParent(dialog_box), wargs, n);

  XtManageChild(dialog_box);
  XtFree(xmstr_mes);
}

/* ----------------------- CCI functions ------------------------ */

/* cci callback function to handle all incoming messages */
cci_do_the_input(caddr_t client_data, int *source, XtInputId *garbage)
{
  MCCIPort s;
  
  s = (MCCIPort) client_data;
  MCCIPoll(s);
}

/* cci lost teacher connection callback */
void cci_lost_teacher(MCCIPort serverPort, void *callBackData)
{
  if (cci_teach_port!=NULL)
    {
      XtRemoveInput(teacher_input_id);
      cci_teach_port = NULL;
    }
}

void cci_lost_student(MCCIPort serverPort, void *callBackData)
{
  int i = (int) callBackData;
  
  XtRemoveInput(student_input_id[i]);  
  cci_student_port[i] = NULL;
}

int get_teach_port()
{ 
  return atoi(teach_port);
}

/* returns the integer value of the port for student student_num */
int get_stu_port(int student_num)
{ 
  return atoi(student_port[student_num]);
}

void cci_anchor_cb(char *url, void *data)
{
  int i;

  /*  printf("*** xwebteach: cci_anchor_cb called with URL: %s\n", url); */
  XmTextSetString(url_field, url);

  if (num_students==0)
    return;
  if (broadcasting==0)
    return;
  for (i=0; i<num_students; i++)
    {
      /* if this student is not connected, try to connect */
      if (cci_student_port[i]==NULL)
	{
	  cci_student_port[i]= (MCCIPort) MCCIConnect(student_host[i], 
						      get_stu_port(i),
						      cci_lost_student, i);
	  if (cci_student_port[i] != NULL)
	    student_input_id[i]=XtAppAddInput(app,
			    MCCIGetSocketDescriptor(cci_student_port[i]),
					      XtInputReadMask,
					      cci_do_the_input,
					      cci_student_port[i]);
	}
      if (cci_student_port[i]!=NULL)  /* if possible, get URL */
	if (MCCIIsConnected(cci_student_port[i]))
	  MCCIGet(cci_student_port[i], url, MCCI_DEFAULT, MCCI_ABSOLUTE, 0);
    }
}

int cci_connect_to_teacher()
/* returns 0 on no error, 1 on error */
{
/*  printf("*** xwebteach: connect to teacher called\n");*/
  if ( (teach_host==NULL) || (teach_port==NULL) )                /* error */
    {
      xwt_message_dialog(form_widget, "Error", no_teach_host_msg, 1);
      XmToggleButtonSetState(toggle, FALSE, TRUE);
      return 1;
    }
  cci_teach_port = (MCCIPort) MCCIConnect(teach_host, get_teach_port(), 
					  cci_lost_teacher, NULL);
  if (!cci_teach_port)                              /* error */
    {
      xwt_message_dialog(form_widget, "Error", no_connect_msg, 1);
      XmToggleButtonSetState(toggle, FALSE, TRUE);
      return 1;
    }
  else
    {
      teacher_input_id=XtAppAddInput(app,
				     MCCIGetSocketDescriptor(cci_teach_port),
				     XtInputReadMask,
				     cci_do_the_input, cci_teach_port);
      return 0;
    }
} /* end of cci_connect_to_teacher()  */

void cci_update_student_connections()
{
  int i;
  
  if (num_students==0)
    return;
  for (i=0; i<num_students; i++)  /* for each student, check connection */
    {
      if (cci_student_port[i]==NULL)
	{
	  cci_student_port[i]= (MCCIPort) MCCIConnect(student_host[i], 
						      get_stu_port(i), 
						      cci_lost_student, i);
	  if (cci_student_port[i]!=NULL)
	    student_input_id[i]=XtAppAddInput(app,
			   MCCIGetSocketDescriptor(cci_student_port[i]),
					   XtInputReadMask, cci_do_the_input, 
					   cci_student_port[i]);
	}
    }
} /* of  cci_update_student_connections() */

void cci_kill_student_connections()
{
  int i;
  
  if (num_students==0)
    return;
  for (i=0; i<num_students; i++)
    if (cci_student_port[i]!=NULL)
      {
	XtRemoveInput(student_input_id[i]);
	MCCIDisconnect(cci_student_port[i]);
      }
}

/* ----------------------- broadcast toggle callback ---------------- */
void toggle_cb(Widget w, caddr_t client_data, 
	       XmToggleButtonCallbackStruct *call_data)
{
/*  printf("*** xwebteach: toggle callback called, state: %d\n", call_data->set); */
  if (call_data->set==True)
    {
      broadcasting=1;
      if (cci_teach_port==NULL)      /* if no teacher connection exists, */
	cci_connect_to_teacher();                 /* attempt to make one */
      if (cci_teach_port!=NULL)
	{
	  /* send a SEND ANCHOR BEGIN command to teacher machine */
	  if (MCCIIsConnected(cci_teach_port))
	    MCCISendAnchor(cci_teach_port, 1, cci_anchor_cb, 0);
	  
	  /* update connections to all student machines */
	  cci_update_student_connections();
	}
    }
  else
    {
      /* send a SEND ANCHOR STOP command to teacher machine */
      if (cci_teach_port!=NULL)
	if (MCCIIsConnected(cci_teach_port))
	  MCCISendAnchor(cci_teach_port, 0, cci_anchor_cb, 0);
      broadcasting=0;
    }
/*  printf("*** xwebteach: broadcasting: %d\n", broadcasting);*/
  return;
} /* end of toggle_cb()  */

/* ---------------------- read in config file --------------------- */

char *get_second_word(char s[300])
{
  char *p, full_string[300];

  strcpy(full_string, s);
  p=strtok(full_string, " ");
  p=strtok(NULL, " ");
  if (p==NULL)
    return NULL;
  else
    return strdup(p);
}

char *get_third_word(char s[300])
{
  char *p, full_string[300];

  strcpy(full_string, s);
  p=strtok(full_string, " ");
  p=strtok(NULL, " ");
  p=strtok(NULL, " ");
  if (p==NULL)
    return NULL;
  else
    return strdup(p);
}

xwt_process_config_str(char config_str[300])
{
  char *dummy_string;

  if (strncmp(config_str, "#", 1)==0)       /* if this line is a comment */
    return;
  if (strncmp(config_str, "teacher", 7)==0)
    {
      if (teach_host!=NULL)
	free(teach_host);
      if (teach_port!=NULL)
	free(teach_port);
      if ((teach_host=get_second_word(config_str))==NULL)
	return;
      if ((teach_port=get_third_word(config_str))==NULL)
	return;
    }
  if (strncmp(config_str, "student", 7)==0)
    {
      if (num_students>=MAX_STUDENTS)
	{
	  printf("\nMaximum number of students reached.\n");
	  return;
	}
      if ((student_host[num_students]=get_second_word(config_str))==NULL)
	return;
      if ((student_port[num_students]=get_third_word(config_str))==NULL)
	return;
      num_students++;
    }
}

int xwt_read_config(int argc, char *argv[]) 
/* returns 0 if no error, returns 1 on error */
{
  int x=1;
  char config_filename[256], next_line[300];
  FILE *config_fileptr;

  if (argc==1)
    return 0;
  do                                        /* check command-line flags */
    {
      if (strcmp(argv[x], "-config")==0)
	{
	  if (argc==x+1)
	    return 1;
	  strcpy(config_filename, argv[x+1]);    /* read in config file */
	  if ( (config_fileptr=fopen(config_filename, "r"))!=NULL )
	    {
	      while ( fgets(next_line, 300, config_fileptr)!=NULL )
		xwt_process_config_str(next_line);
	      fclose(config_fileptr);
	    }
	  else
	    printf("\nConfig file %s unavailable for reading.\n", 
		   config_filename); 
	} 
    }  while (++x<argc);
  return 0;
}

/* -------------------- purge student list --------------------- */
void xwt_purge_student_list()
{
  int x=0;

  if (num_students==0)
    return;
  for (x=0; x<num_students; x++)
    {
      current_selection= -1;
      free(student_host[x]);
      free(student_port[x]);
      if (student_input_id[x]!=NULL)
	XtRemoveInput(student_input_id[x]);
      MCCIDisconnect(cci_student_port[x]);
      cci_student_port[x]=NULL;
    }
  num_students=0;
}

/* ---------------------- updates -------------------------------- */ 
xwt_update_list()
/* make an XmString, xmstr_student_list, for the XmList widget */
{
  int n=0;

  if (num_students==0)
    return;
  if (xmstr_student_list!=NULL)
    XtFree(xmstr_student_list);
  xmstr_student_list = (XmString *) XtMalloc(sizeof(XmString) * num_students);
  for (n=0; n<num_students; n++)
    {
      strcpy(dummy, student_host[n]);    /* build one string from the two */
      strcat(dummy, ":");
      strcat(dummy, student_port[n]);
      xmstr_student_list[n]=XmStringCreate(dummy, XmSTRING_DEFAULT_CHARSET);
    }
  if (broadcasting==1)
    cci_update_student_connections();
}

/* ------------------- check geometry of main window ---------------- */
/* -------- and set the geometry of pop-up widget w accordingly ----- */
void geo_request(Widget w)
{
  Arg wargs[3];
  int n=0;

  XtSetArg(wargs[n], XmNx, &geo_x); n++;
  XtSetArg(wargs[n], XmNy, &geo_y); n++;
  XtGetValues(toplevel, wargs, n);
  n=0;
  XtSetArg(wargs[n], XmNx, (geo_x+30)); n++;
  XtSetArg(wargs[n], XmNy, (geo_y+40)); n++;
  XtSetValues(w, wargs, n);
}

/* -----------------------  redisplay list widget -------------- */
xwt_redisplay_list()
{
  Arg wargs[3];
  int n=0;

  if (teach_host != NULL)
    {
      strcpy(dummy, teach_host);  /* redisplay teacher */
      if (teach_port != NULL)
	{
	  strcat(dummy, ":");
	  strcat(dummy, teach_port);
	}
      XmTextSetString(teacher_field, dummy);
    }
  xwt_update_list();          /* redisplay student list */
  XtUnmanageChild(student_list);
  XtSetArg(wargs[n], XmNitems, xmstr_student_list); n++;
  XtSetArg(wargs[n], XmNitemCount, num_students); n++;
  XtSetValues(student_list, wargs, n);
  XtManageChild(student_list);
  current_selection= -1;
}

/* ---------------------------------------------------------------- */
/* ---------------------- callbacks ------------------------------- */
/* ---------------------------------------------------------------- */
void help_cb(Widget w, caddr_t client_data, caddr_t call_data)
{
  xwt_message_dialog(form_widget, "Help", help_message, 1);
  return;
}

void quit_cb(Widget w, caddr_t client_data, caddr_t call_data)
{
  if (cci_teach_port != NULL)
    {
      MCCIDisconnect(cci_teach_port);
      XtRemoveInput(teacher_input_id);
    }
  cci_kill_student_connections();
  XtCloseDisplay(XtDisplay(w));
  exit(0);
}

/* ---------------------- setup dialog and callbacks -------------------- */

void close_setup_window()
{
  char *setup_str;

  setup_str = XmTextGetString(setup_host_field);  /* redisplay teach host */
  if (teach_host!=NULL)
    free(teach_host);
  teach_host=strdup(setup_str);
  XtFree(setup_str);
  setup_str=XmTextGetString(setup_port_field);  /* redisplay teach port */
  if (teach_port!=NULL)
    free(teach_port);
  teach_port=strdup(setup_str);
  XtFree(setup_str);
  strcpy(dummy, teach_host);                 /* print info in main window */
  strcat(dummy, ":");
  strcat(dummy, teach_port);
  XmTextSetString(teacher_field, dummy);

  xwt_grey_out_all(TRUE);
  XtCallbackPopdown(setup_parent, &setup_idrec, NULL); /* XtPopdown */

  /* if teach host or teach port have changed, break old connection */
  if (cci_teach_port != NULL)
    if ( (strcmp(old_teacher_host_string, teach_host) !=0) ||
	 (strcmp(old_teacher_port_string, teach_port) !=0) )
      {
	MCCIDisconnect(cci_teach_port);
	XtRemoveInput(teacher_input_id);
	cci_teach_port=NULL;
	XmToggleButtonSetState(toggle, FALSE, TRUE);      
      }
  return;
}

void load_cancel_cb(Widget w, caddr_t client_data, caddr_t call_data)
{
  XtCallbackPopdown(load_dialog, &load_idrec, NULL); /* XtPopdown */
  return;
} 

void save_cancel_cb(Widget w, caddr_t client_data, caddr_t call_data)
{
  XtCallbackPopdown(save_dialog, &save_idrec, NULL); /* XtPopdown */
  return;
}

void load_ok_cb(Widget w, caddr_t client_data, 
		XmSelectionBoxCallbackStruct *call_data)
/* load a config file into memory */
{
  char *filename, next_line[300];
  FILE *load_file_ptr;

  XmStringGetLtoR(call_data->value, XmSTRING_DEFAULT_CHARSET, &filename);
  XtCallbackPopdown(load_dialog, &load_idrec, NULL); /* XtPopdown */
  xwt_purge_student_list();
  close_setup_window();
  if ((load_file_ptr=fopen(filename, "r"))!=NULL)
    {
      while (fgets(next_line, 300, load_file_ptr)!=NULL)
	xwt_process_config_str(next_line);
      fclose(load_file_ptr);
    }
  else
    printf("\nConfig file %s unavailable for reading.\n", filename);
  xwt_redisplay_list();
  current_selection= -1;
  return;
}  /* end of load_ok_cb()   */

void output_save_file(FILE *f)
{
  int x;

  /* first write some comments */
  fputs("### X Web Teach   version ", f);
  fputs(VERSION, f);
  fputs("\n### configuration file\n\n", f);
  /* write teacher hostname and port */
  fputs("### teacher hostname and port\n", f);
  fputs("teacher: ", f);
  fputs(teach_host, f);
  fputs(" ", f);
  fputs(teach_port, f);

  /* write all students */
  fputs("\n\n### student hostnames and ports\n", f);
  for (x=0; x<num_students; x++)
    {
      fputs("student: ", f);
      fputs(student_host[x], f);
      fputs(" ", f);
      fputs(student_port[x], f);
      fputs("\n", f);
    }
} /* end of output_save_file ()  */

void save_ok_cb(Widget w, caddr_t client_data, 
		XmSelectionBoxCallbackStruct *call_data)
/* write the current teacher/student list to a config file */
{
  char *filename;
  FILE *save_file_ptr;

  XmStringGetLtoR(call_data->value, XmSTRING_DEFAULT_CHARSET, &filename);
  XtCallbackPopdown(save_dialog, &save_idrec, NULL); /* XtPopdown */
  close_setup_window();
  if ( (save_file_ptr=fopen(filename, "w"))!=NULL)
    {
      output_save_file(save_file_ptr);
      fclose(save_file_ptr);
    }
  else
    printf("\nWriting config file %s failed.\n", filename);
  return;
}

void load_cb(Widget w, caddr_t client_data, caddr_t call_data)
{
  geo_request(load_parent_widget);
  XtCallbackNone(w, load_parent_widget, NULL);  /* used in place of XtPopup */
  load_idrec.shell_widget=load_parent_widget;  /* used by XtCallbackPopdown */
  load_idrec.enable_widget=w;
}

void save_cb(Widget w, caddr_t client_data, caddr_t call_data)
{
  geo_request(save_parent_widget);
  XtCallbackNone(w, save_parent_widget, NULL);  /* used in place of XtPopup */
  save_idrec.shell_widget=save_parent_widget;  /* used by XtCallbackPopdown */
  save_idrec.enable_widget=w;
}

void setup_ok_cb(Widget w, caddr_t client_data, caddr_t call_data)
{
  close_setup_window();
}

void setup_cancel_cb(Widget w, caddr_t client_data, caddr_t call_data)
{
  xwt_grey_out_all(TRUE);  
  XtCallbackPopdown(setup_form, &setup_idrec, NULL); /* XtPopdown */
  return;
} /* end of setup_cancel_cb() */

void setup_save_state()
{
  /* save teacher hostname and port number */
  if (teach_host!=NULL)
    old_teacher_host_string=strdup(teach_host);
  if (teach_port!=NULL)
    old_teacher_port_string=strdup(teach_port);
  /* save all student hostnames and port numbers */
  
}

void setup_cb(Widget w, caddr_t client_data, caddr_t call_data)
{
  geo_request(setup_parent);
  setup_save_state();
  xwt_grey_out_all(FALSE);
  XtCallbackNone(w, setup_parent, NULL);  /* used in place of XtPopup */
  setup_idrec.shell_widget=setup_parent;  /* used by XtCallbackPopdown */
  setup_idrec.enable_widget=w;
  XmTextSetString(setup_host_field, teach_host);  /* default text */
  XmTextSetString(setup_port_field, teach_port);
  XtGrabPointer(setup_host_field, True, 
		FocusChangeMask | EnterWindowMask | PointerMotionMask,
		GrabModeAsync, GrabModeAsync, NULL, None, CurrentTime);
  return;
}  /* end of setup_cb() */

init_setup()
{
  Arg     wargs[15];
  int     n=0;

  setup_parent=XtCreatePopupShell("X Web Teach: Setup",
				  transientShellWidgetClass, 
				  toplevel,  NULL, 0);
  load_parent_widget=XtCreatePopupShell("X Web Teach: Load Configuration File",
				  transientShellWidgetClass, 
				  toplevel,  NULL, 0);
  save_parent_widget=XtCreatePopupShell(
				"X Web Teach: Save Current Configuration",
					transientShellWidgetClass, 
					toplevel,  NULL, 0);
  setup_form=XtCreateManagedWidget("Setup", xmFormWidgetClass,
				   setup_parent, NULL, 0); 
  setup_host_field=XtCreateManagedWidget("host", xmTextWidgetClass, setup_form,
				   NULL, 0);
  setup_port_field=XtCreateManagedWidget("port", xmTextWidgetClass, setup_form,
				   NULL, 0);
  setup_host_label=XtCreateManagedWidget("Host", xmLabelWidgetClass, 
					 setup_form, NULL, 0);
  setup_top_label=XtCreateManagedWidget("Teacher Machine", xmLabelWidgetClass, 
					 setup_form, NULL, 0);
  setup_pot_label=XtCreateManagedWidget("Setup Options", xmLabelWidgetClass, 
					 setup_form, NULL, 0);
  setup_port_label=XtCreateManagedWidget("Port", xmLabelWidgetClass,
					 setup_form,  NULL, 0);
  setup_ok_button=XtCreateManagedWidget("OK", xmPushButtonWidgetClass, 
				  setup_form, NULL, 0);
  setup_cancel_button=XtCreateManagedWidget("Cancel", xmPushButtonWidgetClass, 
				  setup_form, NULL, 0);
  load_button=XtCreateManagedWidget("Load Configuration File", 
				    xmPushButtonWidgetClass, 
				    setup_form, NULL, 0);
  save_button=XtCreateManagedWidget("Save Current Configuration", 
				    xmPushButtonWidgetClass, 
				    setup_form, NULL, 0);
  sep = XtCreateManagedWidget("sep", xmSeparatorWidgetClass, setup_form, 
			      NULL, 0);
  /* load dialog */
  load_dialog= XtCreateManagedWidget("Load Config File", 
				     xmFileSelectionBoxWidgetClass,
				     load_parent_widget, NULL, 0);
  XtUnmanageChild(XmSelectionBoxGetChild(load_dialog, XmDIALOG_HELP_BUTTON));
  XtAddCallback(load_dialog, XmNokCallback, load_ok_cb, NULL);
  XtAddCallback(load_dialog, XmNcancelCallback, load_cancel_cb, NULL);
  /* save dialog */
  save_dialog=XtCreateManagedWidget("Save Config File", 
				    xmFileSelectionBoxWidgetClass,
				    save_parent_widget, NULL, 0);
  XtUnmanageChild(XmSelectionBoxGetChild(save_dialog, XmDIALOG_HELP_BUTTON));
  XtAddCallback(save_dialog, XmNokCallback, save_ok_cb, NULL);
  XtAddCallback(save_dialog, XmNcancelCallback, save_cancel_cb, NULL);
  n=0;
  XtSetArg(wargs[n], XmNtransientFor, form_widget); n++;
  XtSetValues(setup_form, wargs, n);
  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNleftPosition, 1); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 67); n++;
  XtSetArg(wargs[n], XmNrightPosition, 18); n++;
  XtSetArg(wargs[n], XmNtopPosition, 55); n++;
  XtSetValues(setup_host_label, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 1); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 54); n++;
  XtSetArg(wargs[n], XmNrightPosition, 50); n++;
  XtSetArg(wargs[n], XmNtopPosition, 46); n++;
  XtSetArg(wargs[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
  XtSetValues(setup_top_label, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 1); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 13); n++;
  XtSetArg(wargs[n], XmNrightPosition, 99); n++;
  XtSetArg(wargs[n], XmNtopPosition, 1); n++;
  XtSetArg(wargs[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
  XtSetValues(setup_pot_label, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 1); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 43); n++;
  XtSetArg(wargs[n], XmNrightPosition, 99); n++;
  XtSetArg(wargs[n], XmNtopPosition, 42); n++;
  XtSetArg(wargs[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
  XtSetValues(sep, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 1); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 24); n++;
  XtSetArg(wargs[n], XmNrightPosition, 99); n++;
  XtSetArg(wargs[n], XmNtopPosition, 13); n++;
  XtSetArg(wargs[n], XmNtraversalOn, False); n++;
  XtSetValues(load_button, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 1); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 37); n++;
  XtSetArg(wargs[n], XmNrightPosition, 99); n++;
  XtSetArg(wargs[n], XmNtopPosition, 25); n++;
  XtSetArg(wargs[n], XmNtraversalOn, False); n++;
  XtSetValues(save_button, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 19); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 67); n++;
  XtSetArg(wargs[n], XmNrightPosition, 99); n++;
  XtSetArg(wargs[n], XmNtopPosition, 55); n++;
  XtSetValues(setup_host_field, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 1); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 80); n++;
  XtSetArg(wargs[n], XmNrightPosition, 18); n++;
  XtSetArg(wargs[n], XmNtopPosition, 68); n++;
  XtSetValues(setup_port_label, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 19); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 80); n++;
  XtSetArg(wargs[n], XmNrightPosition, 99); n++;
  XtSetArg(wargs[n], XmNtopPosition, 68); n++;
  XtSetValues(setup_port_field, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 65); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 97); n++;
  XtSetArg(wargs[n], XmNrightPosition, 90); n++;
  XtSetArg(wargs[n], XmNtopPosition, 85); n++;
  XtSetArg(wargs[n], XmNtraversalOn, False); n++;
  XtSetValues(setup_ok_button, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 10); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 97); n++;
  XtSetArg(wargs[n], XmNrightPosition, 35); n++;
  XtSetArg(wargs[n], XmNtopPosition, 85); n++;
  XtSetArg(wargs[n], XmNtraversalOn, False); n++;
  XtSetValues(setup_cancel_button, wargs, n);

  XtAddCallback(setup_ok_button, XmNactivateCallback, setup_ok_cb,
		NULL); 
  XtAddCallback(setup_cancel_button, XmNactivateCallback, setup_cancel_cb,
		NULL);
  XtAddCallback(load_button, XmNactivateCallback, load_cb, NULL);
  XtAddCallback(save_button, XmNactivateCallback, save_cb, NULL);
  return;
} /* end of init_setup() */

/* ----------------------- remove a student -------------------- */
void remove_cb(Widget w, caddr_t client_data, caddr_t call_data)
{
  Arg wargs[10];
  int x=current_selection, n;

  if (num_students==0)
    {
      xwt_message_dialog(form_widget, "Error", no_students_msg, 1);
      return;
    }
  if (current_selection== -1)
    {
      xwt_message_dialog(form_widget, "Error", no_selected_msg, 1);
      return;
    }
  if (current_edit!= -1)
    return;
  free(student_host[x]);
  free(student_port[x]);
  if (cci_student_port[x]!=NULL)   /* kill connection */
    {
      MCCIDisconnect(cci_student_port[x]);
      XtRemoveInput(student_input_id[x]);
    }
  for (; x<num_students-1; x++)
    {
      student_host[x]=student_host[x+1];
      student_port[x]=student_port[x+1];
      cci_student_port[x]=cci_student_port[x+1];
      student_input_id[x]=student_input_id[x+1];
    }
  num_students--;
  current_selection= -1;
  xwt_redisplay_list();
} /* end of remove_cb()  */

/* ------------------------- edit a student -------------------- */
void edit_ok_cb(Widget w, Widget client_data, caddr_t call_data)
{
  char *edit_str;
  int  edit_error=0;

  free(student_host[current_edit]);
  free(student_port[current_edit]);
  edit_str = XmTextGetString(edit_host_field);  /* get text from fields */
  if (strlen(edit_str)==0)
    edit_error=1;
  student_host[current_edit]=strdup(edit_str);
  XtFree(edit_str);
  edit_str=XmTextGetString(edit_port_field);
  if (strlen(edit_str)==0)
    edit_error=2;
  student_port[current_edit]=strdup(edit_str);
  XtFree(edit_str);
  if (edit_error==0)
    {
      if (cci_student_port[current_edit]!=NULL)
	{
	  MCCIDisconnect(cci_student_port[current_edit]);
	  XtRemoveInput(student_input_id[current_edit]);
	}
      cci_student_port[current_edit]=NULL;
      xwt_redisplay_list();
      XtCallbackPopdown(edit_parent, &edit_idrec, NULL); /* XtPopdown */
      current_edit= -1;
      xwt_grey_out_all(TRUE);
    }
  else
    xwt_message_dialog(form_widget, "Error", add_error_msg, 1);
  return;
} /* end of edit_ok_cb() */

void edit_cancel_cb(Widget w, caddr_t client_data, caddr_t call_data)
{
  XtCallbackPopdown(edit_form, &edit_idrec, NULL); /* XtPopdown */
  current_edit= -1;
  xwt_grey_out_all(TRUE);
  return;
} /* end of edit_cancel_cb() */

void edit_cb(Widget w, caddr_t client_data, caddr_t call_data)
{
  current_edit=current_selection;
  if (num_students==0)
    {
      xwt_message_dialog(form_widget, "Error", no_students_msg, 1);
      return;
    }
  if (current_selection== -1)
    {
      xwt_message_dialog(form_widget, "Error", no_selected_msg, 1);
      return;
    }
  geo_request(edit_parent);  
  xwt_grey_out_all(FALSE);
  XtCallbackNone(w, edit_parent, NULL);  /* used in place of XtPopup */
  edit_idrec.shell_widget=edit_parent;  /* used by XtCallbackPopdown */
  edit_idrec.enable_widget=w;
  XmTextSetString(edit_host_field, student_host[current_edit]);
  XmTextSetString(edit_port_field, student_port[current_edit]);
  XtGrabPointer(edit_host_field, True, 
		FocusChangeMask | EnterWindowMask | PointerMotionMask,
		GrabModeAsync, GrabModeAsync, NULL, None, CurrentTime);
  return;
}  /* end of edit_cb() */

init_edit()
{
  Arg     wargs[15];
  int     n=0;

  edit_parent=XtCreatePopupShell("X Web Teach: Edit Student Machine",
				  transientShellWidgetClass, 
				  toplevel,  NULL, 0);
  
  edit_form=XtCreateManagedWidget("Edit", xmFormWidgetClass,
				   edit_parent, NULL, 0); 
  edit_host_field=XtCreateManagedWidget("host", xmTextWidgetClass, edit_form,
				   NULL, 0);
  edit_port_field=XtCreateManagedWidget("port", xmTextWidgetClass, edit_form,
				   NULL, 0);
  edit_host_label=XtCreateManagedWidget("Host", xmLabelWidgetClass, 
					 edit_form, NULL, 0);
  edit_port_label=XtCreateManagedWidget("Port", xmLabelWidgetClass,
					 edit_form,  NULL, 0);
  edit_ok_button=XtCreateManagedWidget("OK", xmPushButtonWidgetClass, 
				  edit_form, NULL, 0);
  edit_cancel_button=XtCreateManagedWidget("Cancel", xmPushButtonWidgetClass, 
				  edit_form, NULL, 0);

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNleftPosition, 1); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 30); n++;
  XtSetArg(wargs[n], XmNrightPosition, 24); n++;
  XtSetArg(wargs[n], XmNtopPosition, 5); n++;
  XtSetValues(edit_host_label, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 25); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 30); n++;
  XtSetArg(wargs[n], XmNrightPosition, 99); n++;
  XtSetArg(wargs[n], XmNtopPosition, 5); n++;
  XtSetValues(edit_host_field, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 1); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 60); n++;
  XtSetArg(wargs[n], XmNrightPosition, 24); n++;
  XtSetArg(wargs[n], XmNtopPosition, 35); n++;
  XtSetValues(edit_port_label, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 25); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 60); n++;
  XtSetArg(wargs[n], XmNrightPosition, 99); n++;
  XtSetArg(wargs[n], XmNtopPosition, 35); n++;
  XtSetValues(edit_port_field, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 65); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 90); n++;
  XtSetArg(wargs[n], XmNrightPosition, 90); n++;
  XtSetArg(wargs[n], XmNtopPosition, 65); n++;
  XtSetArg(wargs[n], XmNtraversalOn, False); n++;
  XtSetValues(edit_ok_button, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 10); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 90); n++;
  XtSetArg(wargs[n], XmNrightPosition, 35); n++;
  XtSetArg(wargs[n], XmNtopPosition, 65); n++;
  XtSetArg(wargs[n], XmNtraversalOn, False); n++;
  XtSetValues(edit_cancel_button, wargs, n);

  XtAddCallback(edit_ok_button, XmNactivateCallback, edit_ok_cb,
		NULL); 
  XtAddCallback(edit_cancel_button, XmNactivateCallback, edit_cancel_cb,
		NULL);
  return;
} /* end of init_edit() */

/* -------------------- add a student ------------------- */
int in_student_list(char *check_name, char *check_port)
/* return 1 if in the list, return 0 otherwise */
{
  int x;

  for (x=0; x<num_students; x++)
    if (strcmp(student_host[x], check_name)==0)
      if (strcmp(student_port[x], check_port)==0)
	return 1;
  return 0;
}

void add_ok_cb(Widget w, Widget client_data, caddr_t call_data)
{
  char *add_str;
  int  add_error=0;
  Arg  wargs[2];
  int  n=0;

  add_str = XmTextGetString(add_host_field);  /* get text from fields */
  if (strlen(add_str)==0)
    add_error=1;
  student_host[num_students]=strdup(add_str);
  XtFree(add_str);
  add_str=XmTextGetString(add_port_field);
  if (strlen(add_str)==0)
    add_error=2;
  student_port[num_students]=strdup(add_str);
  XtFree(add_str);
  if (in_student_list(student_host[num_students],  
		      student_port[num_students])==1)
    {
      xwt_message_dialog(form_widget, "Error", already_used_msg, 1);
      return;
    }
  if (add_error==0)
    {
      num_students++;
      XtSetArg(wargs[n], XmNselectedItemCount, 0); n++;
      XtSetValues(student_list, wargs, n);
      xwt_redisplay_list();
      XtCallbackPopdown(add_parent, &add_idrec, NULL); /* XtPopdown */
      xwt_grey_out_all(TRUE);
    }
  else
    xwt_message_dialog(form_widget, "Error", add_error_msg, 1);
  return;
} /* end of add_ok_cb() */

void add_cancel_cb(Widget w, caddr_t client_data, caddr_t call_data)
{
  XtCallbackPopdown(add_form, &add_idrec, NULL); /* XtPopdown */
  xwt_grey_out_all(TRUE);  
  return;
} /* end of add_cancel_cb() */

void add_cb(Widget w, caddr_t client_data, caddr_t call_data)
{
  if (num_students>=MAX_STUDENTS)
    return;
  geo_request(add_parent);
  xwt_grey_out_all(FALSE);
  XtCallbackNone(w, add_parent, NULL);  /* used in place of XtPopup */
  add_idrec.shell_widget=add_parent;  /* used by XtCallbackPopdown */
  add_idrec.enable_widget=w;
  XmTextSetString(add_host_field, "");  /* default text */
  XmTextSetString(add_port_field, "");
  XtGrabPointer(add_host_field, True,
		FocusChangeMask | EnterWindowMask | PointerMotionMask,
		GrabModeAsync, GrabModeAsync, NULL, None, CurrentTime);
  return;
}  /* end of add_cb() */

init_add()
{
  Arg     wargs[15];
  int     n=0;

  add_parent=XtCreatePopupShell("X Web Teach: Add Student Machine",
				  transientShellWidgetClass, 
				  form_widget,  NULL, 0);
  add_form=XtCreateManagedWidget("Add", xmFormWidgetClass,
				   add_parent, NULL, 0); 
  add_host_field=XtCreateManagedWidget("host", xmTextWidgetClass, add_form,
				   NULL, 0);
  add_port_field=XtCreateManagedWidget("port", xmTextWidgetClass, add_form,
				   NULL, 0);
  add_host_label=XtCreateManagedWidget("Host", xmLabelWidgetClass, 
					 add_form, NULL, 0);
  add_port_label=XtCreateManagedWidget("Port", xmLabelWidgetClass,
					 add_form,  NULL, 0);
  add_ok_button=XtCreateManagedWidget("OK", xmPushButtonWidgetClass, 
				  add_form, NULL, 0);
  add_cancel_button=XtCreateManagedWidget("Cancel", xmPushButtonWidgetClass, 
				  add_form, NULL, 0);

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNleftPosition, 1); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 30); n++;
  XtSetArg(wargs[n], XmNrightPosition, 24); n++;
  XtSetArg(wargs[n], XmNtopPosition, 5); n++;
  XtSetValues(add_host_label, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 25); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 30); n++;
  XtSetArg(wargs[n], XmNrightPosition, 99); n++;
  XtSetArg(wargs[n], XmNtopPosition, 5); n++;
  XtSetValues(add_host_field, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 1); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 60); n++;
  XtSetArg(wargs[n], XmNrightPosition, 24); n++;
  XtSetArg(wargs[n], XmNtopPosition, 35); n++;
  XtSetValues(add_port_label, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 25); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 60); n++;
  XtSetArg(wargs[n], XmNrightPosition, 99); n++;
  XtSetArg(wargs[n], XmNtopPosition, 35); n++;
  XtSetValues(add_port_field, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 65); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 90); n++;
  XtSetArg(wargs[n], XmNrightPosition, 90); n++;
  XtSetArg(wargs[n], XmNtopPosition, 65); n++;
  XtSetArg(wargs[n], XmNtraversalOn, False); n++;
  XtSetValues(add_ok_button, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 10); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 90); n++;
  XtSetArg(wargs[n], XmNrightPosition, 35); n++;
  XtSetArg(wargs[n], XmNtopPosition, 65); n++;
  XtSetArg(wargs[n], XmNtraversalOn, False); n++;
  XtSetValues(add_cancel_button, wargs, n);

  XtAddCallback(add_ok_button, XmNactivateCallback, add_ok_cb,
		NULL); 
  XtAddCallback(add_cancel_button, XmNactivateCallback, add_cancel_cb,
		NULL);
  return;
} /* end of init_add() */

/* --------------------------- student selection -------------------- */
void list_cb(Widget w, caddr_t client_data, XmListCallbackStruct *call_data)
{
  if (call_data->selected_item_count==0)
    current_selection= -1;
  else
    current_selection=(call_data->item_position)-1;
  return;
}  /* end of list_cb() */

/* ------------------- initialization of gui ----------------------------- */
xwt_gui()  /* create widgets and their attachments for the GUI */
{
  int n;
  Arg wargs[15];
  XmString str;

  /* main container form widget */
  form_widget = XtCreateManagedWidget("form_widget", 
				      xmFormWidgetClass, 
				      toplevel, NULL, 0);
  /* teacher, title, URL text field */
  toggle = XtCreateManagedWidget("Broadcasting", xmToggleButtonWidgetClass,
				 form_widget, NULL, 0);
  teacher_field = XtCreateManagedWidget("Teacher", xmTextWidgetClass, 
					form_widget, NULL, 0);
/*
  title_field = XtCreateManagedWidget("Title", xmTextWidgetClass, 
				      form_widget, NULL, 0);
*/
  url_field = XtCreateManagedWidget("URL", xmTextWidgetClass, 
				    form_widget, NULL, 0);
  /* all the buttons */
  setup_button=XtCreateManagedWidget("Setup", 
				      xmPushButtonWidgetClass, 
				      form_widget, NULL, 0);
  edit_button=XtCreateManagedWidget("Edit", 
				      xmPushButtonWidgetClass, 
				      form_widget, NULL, 0);

  help_button=XtCreateManagedWidget("Help", 
				      xmPushButtonWidgetClass, 
				      form_widget, NULL, 0);
  quit_button=XtCreateManagedWidget("Quit", 
				      xmPushButtonWidgetClass, 
				      form_widget, NULL, 0);
  add_button=XtCreateManagedWidget("Add", 
				      xmPushButtonWidgetClass, 
				      form_widget, NULL, 0);
  remove_button=XtCreateManagedWidget("Remove", 
				      xmPushButtonWidgetClass, 
				      form_widget, NULL, 0);
  /* gui labels */
  teacher_label=XtCreateManagedWidget("Teacher", 
				      xmLabelWidgetClass, 
				      form_widget, NULL, 0);
  student_label=XtCreateManagedWidget("Students", 
				      xmLabelWidgetClass, 
				      form_widget, NULL, 0);
/*
  title_label=XtCreateManagedWidget("Title", 
				    xmLabelWidgetClass, 
				    form_widget, NULL, 0);
*/
  url_label=XtCreateManagedWidget("URL", 
				  xmLabelWidgetClass, 
				  form_widget, NULL, 0);
  /* set the resource attachments to place all the widgets */
  n=0;
  XtSetArg(wargs[n], XmNkeyboardFocusPolicy, XmPOINTER); n++;
  XtSetArg(wargs[n], XmNhighlightThickness, 0); n++;
  XtSetArg(wargs[n], XmNtitle, "X Web Teach"); n++;
  XtSetValues(toplevel, wargs, n);
  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_POSITION); n++;
  XtSetArg(wargs[n], XmNleftPosition, 34); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 10); n++;
  XtSetArg(wargs[n], XmNrightPosition, 66); n++;
  XtSetArg(wargs[n], XmNtopPosition, 1); n++;
  XtSetValues(help_button, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 68); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 10); n++;
  XtSetArg(wargs[n], XmNrightPosition, 99); n++;
  XtSetArg(wargs[n], XmNtopPosition, 1); n++;
  XtSetValues(quit_button, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 1); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 40); n++;
  XtSetArg(wargs[n], XmNrightPosition, 99); n++;
  XtSetArg(wargs[n], XmNtopPosition, 31); n++;
  XtSetArg(wargs[n], XmNeditable, FALSE); n++;
  XtSetArg(wargs[n], XmNcursorPositionVisible, FALSE); n++;
  XtSetValues(teacher_field, wargs, n);
  n=4; 
  XtSetArg(wargs[n], XmNleftPosition, 55); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 30); n++;
  XtSetArg(wargs[n], XmNrightPosition, 98); n++;
  XtSetArg(wargs[n], XmNtopPosition, 21); n++;
  str = XmStringCreateLtoR("Broadcasting", XmSTRING_DEFAULT_CHARSET);
  XtSetArg(wargs[n], XmNlabelString, str); n++;
  XtSetArg(wargs[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
  XtSetArg(wargs[n], XmNfillOnSelect, TRUE); n++;
  XtSetValues(toggle, wargs, n);
  XmStringFree(str);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 1); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 10); n++;
  XtSetArg(wargs[n], XmNrightPosition, 32); n++;
  XtSetArg(wargs[n], XmNtopPosition, 1); n++;
  XtSetValues(setup_button, wargs, n);
/*
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 21); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 18); n++;
  XtSetArg(wargs[n], XmNrightPosition, 99); n++;
  XtSetArg(wargs[n], XmNtopPosition, 10); n++;
  XtSetArg(wargs[n], XmNeditable, FALSE); n++;
  XtSetArg(wargs[n], XmNcursorPositionVisible, FALSE); n++;
  XtSetValues(title_field, wargs, n);
*/
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 21); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 20); n++;
  XtSetArg(wargs[n], XmNrightPosition, 99); n++;
  XtSetArg(wargs[n], XmNtopPosition, 11); n++;
  XtSetArg(wargs[n], XmNeditable, FALSE); n++;
  XtSetArg(wargs[n], XmNcursorPositionVisible, FALSE); n++;
  XtSetValues(url_field, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 1); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 89); n++;
  XtSetArg(wargs[n], XmNrightPosition, 99); n++;
  XtSetArg(wargs[n], XmNtopPosition, 51); n++;
  XtSetArg(wargs[n], XmNselectionPolicy, XmSINGLE_SELECT); n++;
  XtSetArg(wargs[n], XmNitems, xmstr_student_list); n++;
  XtSetArg(wargs[n], XmNitemCount, num_students); n++;
  student_list = XmCreateScrolledList(form_widget, "Students", wargs, n); 
  XtManageChild(student_list);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 1); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 99); n++;
  XtSetArg(wargs[n], XmNrightPosition, 32); n++;
  XtSetArg(wargs[n], XmNtopPosition, 90); n++;
  XtSetValues(add_button, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 34); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 99); n++;
  XtSetArg(wargs[n], XmNrightPosition, 66); n++;
  XtSetArg(wargs[n], XmNtopPosition, 90); n++;
  XtSetValues(remove_button, wargs, n);
  n=4;
  XtSetArg(wargs[n], XmNleftPosition, 68); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 99); n++;
  XtSetArg(wargs[n], XmNrightPosition, 99); n++;
  XtSetArg(wargs[n], XmNtopPosition, 90); n++;
  XtSetValues(edit_button, wargs, n);
  /* form widget text labels */
  n=4;
  str = XmStringCreateLtoR("Teacher", XmSTRING_DEFAULT_CHARSET);
  XtSetArg(wargs[n], XmNlabelString, str); n++;
  XtSetArg(wargs[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
  XtSetArg(wargs[n], XmNleftPosition, 1); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 30); n++;
  XtSetArg(wargs[n], XmNrightPosition, 32); n++;
  XtSetArg(wargs[n], XmNtopPosition, 24); n++;
  XtSetValues(teacher_label, wargs, n); 
  XmStringFree(str);
  n=4;  
  str = XmStringCreateLtoR("Students", XmSTRING_DEFAULT_CHARSET);
  XtSetArg(wargs[n], XmNlabelString, str); n++;
  XtSetArg(wargs[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
  XtSetArg(wargs[n], XmNleftPosition, 1); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 50); n++;
  XtSetArg(wargs[n], XmNrightPosition, 25); n++;
  XtSetArg(wargs[n], XmNtopPosition, 44); n++;
  XtSetValues(student_label, wargs, n); 
  XmStringFree(str);
  n=4;  
/*
  str = XmStringCreateLtoR("Title", XmSTRING_DEFAULT_CHARSET);
  XtSetArg(wargs[n], XmNlabelString, str); n++;
  XtSetArg(wargs[n], XmNalignment, XmALIGNMENT_END); n++;
  XtSetArg(wargs[n], XmNleftPosition, 1); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 18); n++;
  XtSetArg(wargs[n], XmNrightPosition, 20); n++;
  XtSetArg(wargs[n], XmNtopPosition, 13); n++;
  XtSetValues(title_label, wargs, n); 
  XmStringFree(str);
*/
  n=4;  
  str = XmStringCreateLtoR("URL", XmSTRING_DEFAULT_CHARSET);
  XtSetArg(wargs[n], XmNlabelString, str); n++;
  XtSetArg(wargs[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
  XtSetArg(wargs[n], XmNleftPosition, 1); n++;
  XtSetArg(wargs[n], XmNbottomPosition, 20); n++;
  XtSetArg(wargs[n], XmNrightPosition, 20); n++;
  XtSetArg(wargs[n], XmNtopPosition, 11); n++;
  XtSetValues(url_label, wargs, n); 
  XmStringFree(str);
}  /*   end of xwt_gui()  */

xwt_setup_callbacks()  /* register widget callbacks */
{
  XtAddCallback(help_button, XmNactivateCallback, help_cb, NULL);
  XtAddCallback(quit_button, XmNactivateCallback, quit_cb, NULL);
  XtAddCallback(setup_button, XmNactivateCallback, setup_cb, NULL);
  XtAddCallback(add_button, XmNactivateCallback, add_cb, NULL);
  XtAddCallback(remove_button, XmNactivateCallback, remove_cb, NULL);
  XtAddCallback(edit_button, XmNactivateCallback, edit_cb, NULL);
  XtAddCallback(student_list, XmNsingleSelectionCallback, list_cb, NULL); 
  XtAddCallback(toggle, XmNvalueChangedCallback, toggle_cb, NULL);
} /*  end of xwt_setup_callbacks()   */

/* --------------------- main ------------------------ */
main(int argc, char *argv[]) 
{
/*  char *default_resources[]; */
  static XrmOptionDescRec options[] = {" "};

  toplevel = XtAppInitialize(&app, "XWebTeach", options, 0, &argc, argv,
			     default_resources, NULL, 0);
  if (xwt_read_config(argc, argv)==1)  /* parse command line */
    {
      printf("\nUsage: %s -config <filename>\n", argv[0]);
      exit(1);
    }
  xwt_update_list();                 /* generate student list */
  xwt_gui();                         /* init the gui */
  init_setup();
  init_edit();
  init_add();
  xwt_setup_callbacks();             /* setup callback functions */
  XtRealizeWidget(toplevel);
  xwt_redisplay_list();
  MCCIInitialize();
  XtAppMainLoop(app);
}  /* end of main()  */
